from __future__ import annotations

import logging
from typing import Optional, Sequence, Union

from composer.core import Algorithm, Event, State
from composer.loggers import Logger
from composer.utils import module_surgery
from torch.nn import Module
from torch.optim import Optimizer

from flash_attention_softmax_n.surgery.surgery_functions import policy_registry

log = logging.getLogger(__name__)

__all__ = ['AttentionSoftmaxN', 'apply_attention_softmax_n']


def apply_attention_softmax_n(
    model: Module,
    softmax_n_param: float,
    optimizers: Optional[Union[Optimizer, Sequence[Optimizer]]] = None
) -> None:
    """
    Replaces the forward method of SelfAttention with a version that uses softmax_n.

    Example:
    ```python
    import transformers

    from flash_attention_softmax_n.surgery import apply_attention_softmax_n


    model = transformers.AutoModel.from_pretrained('bert-base-uncased')
    apply_attention_softmax_n(model=model, softmax_n_param=1.)

    ```

    :param model: Model to transform.
    :param softmax_n_param: The value of n.
    :param optimizers: Existing optimizers that are bound to `model.parameters()`. Omit this parameters if optimizers will be constructed after calling this function.
    """
    def as_replacement_function(surgery_function):

        def replacement_function(module: Module, module_index: int):
            return surgery_function(module, module_index, softmax_n_param=softmax_n_param)

        return replacement_function

    policies = {
        module_class: as_replacement_function(attention_softmax_n_surgery_function)
        for module_class, attention_softmax_n_surgery_function in policy_registry.items()
    }

    replaced_pairs = module_surgery.replace_module_classes(model, optimizers=optimizers, policies=policies)

    count = len(replaced_pairs)
    if count == 0:
        supported_modules = ''.join(sorted(['\n\t' + c.__module__ + '.' + c.__name__ for c in policy_registry.keys()]))
        log.warning(f'AttentionSoftmaxN had no effect on the model! Support for AttentionSoftmaxN surgery '
                    f'is currently limited to the following classes: {supported_modules}')
    else:
        log.info(f'{count} instances of AttentionSoftmaxN added')


class AttentionSoftmaxN(Algorithm):
    """
    Object that applies attention softmax_n in a Mosaic trainer.

    Example:
    ```python
    import composer
    import transformers

    from flash_attention_softmax_n.surgery import AttentionSoftmaxN


    model = transformers.AutoModel.from_pretrained('bert-base-uncased')
    trainer = composer.trainer.Trainer(
        model=model,
        algorithms=[AttentionSoftmaxN(softmax_n_param=1.)]
    )

    ```
    """
    def __init__(self, softmax_n_param: float) -> None:
        self.softmax_n_param = softmax_n_param
        self._applied = False

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}()'

    @staticmethod
    def required_on_load() -> bool:
        return True

    def match(self, event: Event, state: State) -> bool:
        del state  # unused
        return event == Event.INIT and not self._applied

    def apply(self, event: Event, state: State, logger: Logger) -> None:
        del event, logger  # unused
        apply_attention_softmax_n(
            state.model,
            softmax_n_param=self.softmax_n_param,
            optimizers=state.optimizers,
        )
        self._applied = True



from json import dump
from pathlib import Path
from typing import Dict, Any


def save_results(results: Dict[str, Dict[str, Any]], model_name: str):
    results_dir = Path.cwd() / "results"
    results_dir.mkdir(exist_ok=True)

    filepath = results_dir / f"{model_name}.json"

    with filepath.open(mode='w') as fp:
        dump(results, fp, indent=4)



from typing import Tuple, Optional, Union

from torch import Tensor, mean, pow, no_grad


_Dim = Optional[Union[int, Tuple[int]]]


@no_grad()
def central_moment(x: Tensor, k: int, dim: _Dim = None) -> Tensor:
    """
    Moment of a random variable about the random variable's mean.
    """
    return mean(pow(x - mean(x, dim=dim, keepdim=True), k), dim=dim)


@no_grad()
def variance(x: Tensor, dim: _Dim = None) -> Tensor:
    return central_moment(x, 2, dim=dim)


@no_grad()
def standard_deviation(x: Tensor, dim: _Dim = None) -> Tensor:
    return pow(variance(x, dim=dim), 0.5)


@no_grad()
def standardized_moment(x: Tensor, k: int, dim: _Dim = None) -> Tensor:
    """
    The ratio of the kth moment about the mean.
    """
    return central_moment(x, k, dim=dim) / pow(variance(x, dim=dim), k / 2)


@no_grad()
def skewness(x: Tensor, dim: _Dim = None) -> Tensor:
    return standardized_moment(x, 3, dim=dim)


@no_grad()
def kurtosis(x: Tensor, dim: _Dim = None) -> Tensor:
    """
    Excess kurtosis
    """
    return standardized_moment(x, 4, dim=dim) - 3.


def _get_stat_dim(x: Tensor) -> Tuple[int]:
    """
    If no `dim` is specified, assume `x` is a batch of samples and compute the statistic for each sample.
    """
    return tuple(range(1, x.ndim))


@no_grad()
def variance_batch_mean(x: Tensor) -> float:
    """
    Compute the variance for each sample in a batch, and then compute the mean of the variance in the batch.
    """
    var = variance(x, dim=_get_stat_dim(x))
    return mean(var).item()


@no_grad()
def skewness_batch_mean(x: Tensor) -> float:
    """
    Compute the skewness for each sample in a batch, and then compute the mean of the skewness in the batch.
    """
    skew = skewness(x, dim=_get_stat_dim(x))
    return mean(skew).item()


@no_grad()
def kurtosis_batch_mean(x: Tensor) -> float:
    """
    Compute the kurtosis for each sample in a batch, and then compute the mean of the kurtosis in the batch.
    """
    kurt = kurtosis(x, dim=_get_stat_dim(x))
    return mean(kurt).item()



from collections import defaultdict
from functools import partial
from typing import DefaultDict, Tuple, Optional, Set, Dict
from warnings import warn

from torch import Tensor, no_grad, mean
from torch.nn import Module

from flash_attention_softmax_n.analysis.statistics import (
    kurtosis_batch_mean,
    skewness_batch_mean,
    variance_batch_mean,
    kurtosis,
    skewness,
    variance
)


_activation_stat_funcs = {
        'kurtosis': kurtosis_batch_mean,
        'skewness': skewness_batch_mean,
        'variance': variance_batch_mean,
        'mean': lambda x: mean(x).item()
    }


@no_grad()
def save_activations_statistics(
        activations: DefaultDict,
        name: str,
        module: Module,
        inp: Tuple,
        out: Tensor
) -> None:
    """
    PyTorch Forward hook to compute the online average of statistics at each forward pass.
    Mutates specified dict objects with each fwd pass.
    """
    try:
        batch_size = out.shape[0]
        batch_weight = batch_size / (activations[name]['n_samples'] + batch_size)
        activations[name]['n_samples'] += batch_size

        # new_value == (1 - weight) * current_value + weight * update_value
        for stat, func in _activation_stat_funcs.items():
            # Down-weight the current value
            activations[name][stat] *= (1 - batch_weight)
            # Add the weighted new value
            activations[name][stat] += batch_weight * func(out)

    except AttributeError as e:
        warn(f"Unable to compute stats for module {name}. Consider setting or updating `layers_to_save` in `register_activation_hooks`. {e}.", UserWarning)


def _check_name(name: str, layers_to_save: Optional[Set[str]] = None) -> bool:
    return (layers_to_save is None and 'attention.output' in name) or (layers_to_save is not None and name in layers_to_save)


def register_activation_hooks(
        model: Module,
        layers_to_save: Optional[Set[str]] = None
) -> DefaultDict[str, DefaultDict[str, float]]:
    """Registers forward hooks in specified layers.
    Parameters
    ----------
    model:
        PyTorch model
    layers_to_save:
        Module names within ``model`` whose activations we want to save. If None, save all layers

    Returns
    -------
    activations_dict:
        dict of dicts containing activations of specified layers in
        ``layers_to_save``.
    """
    activations_dict = defaultdict(lambda: defaultdict(float))  # Python floats are double-precision, so representing the integer-valued 'n_samples' this way is fine.

    for name, module in model.named_modules():
        if _check_name(name, layers_to_save):
            module.register_forward_hook(
                partial(save_activations_statistics, activations_dict, name)
            )

    return activations_dict


@no_grad()
def compute_weight_statistics(model: Module) -> Dict[str, Dict[str, float]]:
    results = dict()
    for name, param in model.named_parameters():
        results[name] = {
            'n_weights': param.numel(),
            'kurtosis': kurtosis(param).item(),
            'skewness': skewness(param).item(),
            'variance': variance(param).item(),
            'mean': mean(param).item()
        }
    return results



from einops._torch_specific import allow_ops_in_compiled_graph

from flash_attention_softmax_n.core.flash_attn import flash_attention_n
from flash_attention_softmax_n.core.functional import softmax_n, slow_attention_n
try:
    from flash_attention_softmax_n.core.flash_attn_triton import flash_attention_n_triton
    TRITON_INSTALLED = True
except ModuleNotFoundError:
    TRITON_INSTALLED = False


allow_ops_in_compiled_graph()



try:
    from flash_attention_softmax_n.surgery.surgery_functions import _bert
    from flash_attention_softmax_n.surgery.surgery_functions.utils import policy_registry

    __all__ = ['policy_registry']
except ModuleNotFoundError:
    __all__ = []



from math import sqrt
from types import MethodType
from typing import Optional, Tuple

from torch import Tensor, FloatTensor, cat, matmul, tensor, long, arange, einsum
from torch.nn import Module
from transformers.models.bert.modeling_bert import BertSelfAttention
from transformers.models.roberta.modeling_roberta import RobertaSelfAttention

from flash_attention_softmax_n import softmax_n
from flash_attention_softmax_n.surgery.surgery_functions.utils import policy_registry


@policy_registry.register(BertSelfAttention, RobertaSelfAttention)
def bert_attention_converter(module: Module, module_index: int, softmax_n_param: float) -> Module:
    """Adds AttentionSoftmaxN to Bert-style SelfAttention."""
    assert isinstance(module, (BertSelfAttention, RobertaSelfAttention))
    del module_index  # unused
    module.n = softmax_n_param
    setattr(module, 'forward', MethodType(forward, module))
    return module


def forward(
        self,
        hidden_states: Tensor,
        attention_mask: Optional[FloatTensor] = None,
        head_mask: Optional[FloatTensor] = None,
        encoder_hidden_states: Optional[FloatTensor] = None,
        encoder_attention_mask: Optional[FloatTensor] = None,
        past_key_value: Optional[Tuple[Tuple[FloatTensor]]] = None,
        output_attentions: Optional[bool] = False,
) -> Tuple[Tensor]:
    mixed_query_layer = self.query(hidden_states)

    # If this is instantiated as a cross-attention module, the keys
    # and values come from an encoder; the attention mask needs to be
    # such that the encoder's padding tokens are not attended to.
    is_cross_attention = encoder_hidden_states is not None

    if is_cross_attention and past_key_value is not None:
        # reuse k,v, cross_attentions
        key_layer = past_key_value[0]
        value_layer = past_key_value[1]
        attention_mask = encoder_attention_mask
    elif is_cross_attention:
        key_layer = self.transpose_for_scores(self.key(encoder_hidden_states))
        value_layer = self.transpose_for_scores(self.value(encoder_hidden_states))
        attention_mask = encoder_attention_mask
    elif past_key_value is not None:
        key_layer = self.transpose_for_scores(self.key(hidden_states))
        value_layer = self.transpose_for_scores(self.value(hidden_states))
        key_layer = cat([past_key_value[0], key_layer], dim=2)
        value_layer = cat([past_key_value[1], value_layer], dim=2)
    else:
        key_layer = self.transpose_for_scores(self.key(hidden_states))
        value_layer = self.transpose_for_scores(self.value(hidden_states))

    query_layer = self.transpose_for_scores(mixed_query_layer)

    use_cache = past_key_value is not None
    if self.is_decoder:
        # if cross_attention save Tuple(torch.Tensor, torch.Tensor) of all cross attention key/value_states.
        # Additional calls to cross_attention layer can then reuse all cross-attention
        # key/value_states (first "if" case)
        # if uni-directional self-attention (decoder) save Tuple(torch.Tensor, torch.Tensor) of
        # all previous decoder key/value_states. Additional calls to uni-directional self-attention
        # can concat previous decoder key/value_states to current projected key/value_states (third "elif" case)
        # if encoder bidirectional self-attention `past_key_value` is always `None`
        past_key_value = (key_layer, value_layer)

    # Take the dot product between "query" and "key" to get the raw attention scores.
    attention_scores = matmul(query_layer, key_layer.transpose(-1, -2))

    if self.position_embedding_type == "relative_key" or self.position_embedding_type == "relative_key_query":
        query_length, key_length = query_layer.shape[2], key_layer.shape[2]
        if use_cache:
            position_ids_l = tensor(key_length - 1, dtype=long, device=hidden_states.device).view(-1, 1)
        else:
            position_ids_l = arange(query_length, dtype=long, device=hidden_states.device).view(-1, 1)
        position_ids_r = arange(key_length, dtype=long, device=hidden_states.device).view(1, -1)
        distance = position_ids_l - position_ids_r

        positional_embedding = self.distance_embedding(distance + self.max_position_embeddings - 1)
        positional_embedding = positional_embedding.to(dtype=query_layer.dtype)  # fp16 compatibility

        if self.position_embedding_type == "relative_key":
            relative_position_scores = einsum("bhld,lrd->bhlr", query_layer, positional_embedding)
            attention_scores = attention_scores + relative_position_scores
        elif self.position_embedding_type == "relative_key_query":
            relative_position_scores_query = einsum("bhld,lrd->bhlr", query_layer, positional_embedding)
            relative_position_scores_key = einsum("bhrd,lrd->bhlr", key_layer, positional_embedding)
            attention_scores = attention_scores + relative_position_scores_query + relative_position_scores_key

    attention_scores = attention_scores / sqrt(self.attention_head_size)
    if attention_mask is not None:
        # Apply the attention mask is (precomputed for all layers in RobertaModel forward() function)
        attention_scores = attention_scores + attention_mask

    # Normalize the attention scores to probabilities.
    attention_probs = softmax_n(attention_scores, n=self.n, dim=-1)  # *** modified by CWM ***

    # This is actually dropping out entire tokens to attend to, which might
    # seem a bit unusual, but is taken from the original Transformer paper.
    attention_probs = self.dropout(attention_probs)

    # Mask heads if we want to
    if head_mask is not None:
        attention_probs = attention_probs * head_mask

    context_layer = matmul(attention_probs, value_layer)

    context_layer = context_layer.permute(0, 2, 1, 3).contiguous()
    new_context_layer_shape = context_layer.size()[:-2] + (self.all_head_size,)
    context_layer = context_layer.view(new_context_layer_shape)

    outputs = (context_layer, attention_probs) if output_attentions else (context_layer,)

    if self.is_decoder:
        outputs = outputs + (past_key_value,)
    return outputs



import inspect
import logging
from typing import Callable, Dict, Optional, Type

import torch

log = logging.getLogger(__name__)

AttentionSoftmaxNReplacementFunction = Callable[[torch.nn.Module, int, float], Optional[torch.nn.Module]]


class PolicyRegistry(Dict[Type[torch.nn.Module], AttentionSoftmaxNReplacementFunction]):
    """
    A registry mapping for AttentionSoftmaxN surgery.
    """

    def register(self, *modules: Type[torch.nn.Module]) -> Callable[[AttentionSoftmaxNReplacementFunction], AttentionSoftmaxNReplacementFunction]:
        """
        This decorator registers mappings from torch modules to their surgery functions.

        Example:
        ```python
        # Replace slow_attention_0 in MyModel with flash_attention_n.
        import torch

        from flash_attention_n import slow_attention_n, flash_attention_n
        from flash_attention_n.surgery.surgery_functions import policy_registry


        class MyModel(torch.nn.Module):
            def __init__(self):
                super().__init__()
                self.attn = SlowAttention()

            def forward(self, q, k, v):
                return self.attn(q, k, v, softmax_n_param=0.)


        class SlowAttention(torch.nn.Module):
            def __init__(self):
                super().__init__()

            def forward(self, q, k, v):
                return slow_attention_n(q, k, v, softmax_n_param=0.)


        @policy_registry.register(SlowAttention)
        def slow_attention_converter(module: torch.nn.Module, module_index: int, softmax_n_param: float) -> torch.nn.Module:
            assert isinstance(module, SlowAttention)
            del module_index  # unused
            module.n = softmax_n_param
            setattr(module, 'forward', MethodType(forward, module))
            return module


        def forward(self, q: torch.Tensor, k: torch.Tensor, v: torch.Tensor) -> torch.Tensor:
            return flash_attention_n(q, k, v, softmax_n_param=int(self.n))
        ```
        """
        if len(modules) == 0:
            raise ValueError('Registry decoration without any module class inputs has no effect.')

        def _validate_signature(func: Callable):
            # Necessary to enforce that `func` has a valid signature (i.e. is a AttentionSoftmaxNReplacementFunction)
            signature = inspect.signature(func)
            parameters = signature.parameters
            if len(parameters) != 3:
                raise ValueError(f'Each attention softmax N surgery function must accept 2 arguments, {func} accepts {len(parameters)}')
            ((_, module_param), (_, index_param), (softmax_n_name, softmax_n_param)) = parameters.items()
            if module_param.annotation != torch.nn.Module:
                raise TypeError(f'The first argument of attention softmax N surgery function {func} must be of type "torch.nn.Module"')
            if index_param.annotation != int:
                raise TypeError(f'The second argument of attention softmax N surgery function {func} must be of type "int"')
            if softmax_n_param.annotation != float:
                raise TypeError(f'The third argument of attention softmax N surgery function {func} must be of type "float"')
            if softmax_n_name != 'softmax_n_param':
                raise NameError(f'The third argument of function {func} must be named "softmax_n_param"')

        def _register_module(module: Type[torch.nn.Module], func: Callable) -> None:
            if not issubclass(module, torch.nn.Module):
                raise TypeError(f'Module {module.__name__} is not a subclass of `torch.nn.Module`.')
            if module in self:
                raise ValueError(f'An AttentionSoftmaxNReplacementFunction has already been registered for module {module.__name__}.')
            self[module] = func
            return

        def wrapper(func: AttentionSoftmaxNReplacementFunction) -> AttentionSoftmaxNReplacementFunction:
            _validate_signature(func)
            for module in modules:
                _register_module(module, func)
            return func

        return wrapper


# Initialize the policy registry that AttentionSoftmaxN will reference
policy_registry = PolicyRegistry()



try:
    from flash_attention_softmax_n.surgery.attention_softmax_n import AttentionSoftmaxN, apply_attention_softmax_n
    SURGERY_INSTALLED = True
except ModuleNotFoundError:
    SURGERY_INSTALLED = False
